package com.qgs.dc.s7.controller;

import com.qgs.dc.opcua.arg.*;
import com.qgs.dc.opcua.controller.R;
import com.qgs.dc.s7.service.S7Service;
import org.apache.plc4x.java.PlcDriverManager;
import org.apache.plc4x.java.api.PlcConnection;
import org.apache.plc4x.java.api.exceptions.PlcConnectionException;
import org.apache.plc4x.java.api.messages.PlcReadRequest;
import org.apache.plc4x.java.api.messages.PlcReadResponse;
import org.apache.plc4x.java.api.types.PlcResponseCode;
import org.apache.plc4x.java.api.value.PlcValue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.concurrent.CompletableFuture;

@RestController
@RequestMapping("/s7")
public class S7Controller {
    private static final Logger logger = LoggerFactory.getLogger(S7Controller.class);

    @Autowired
    S7Service s7Service;


    @PostMapping("/addThisPlc")
    public R addThisPlc() throws PlcConnectionException {
        Integer plc1 = s7Service.addPlc("plc1", "s7://192.168.0.200");
        return R.ok().put("res",plc1);
    }
    @PostMapping("/getValue")
    public R getValue() throws PlcConnectionException {
        try(PlcConnection conn = s7Service.getConnection("s7://192.168.0.200")) {
            if(conn.isConnected()){
                if(conn.getMetadata().canRead()){
                    try {
                        PlcReadRequest.Builder builder = conn.readRequestBuilder();
                        //builder.addItem("INT_1", "%DB10:4:INT");
                        //builder.addItem("Byte-array", "%DB10:312:BYTE[20]");
                        builder.addItem("STRING", "%DB10:10.0:STRING(20)");
                        PlcReadRequest readRequest = builder.build();
                        //PlcReadResponse response = readRequest.execute().get();
                        CompletableFuture<? extends PlcReadResponse> execute = readRequest.execute();

                        PlcReadResponse response = execute.get();

                        for (String fieldName : response.getFieldNames()) {
                            if(response.getResponseCode(fieldName) == PlcResponseCode.OK) {
                                int numValues = response.getNumberOfValues(fieldName);
                                PlcValue asPlcValue = response.getAsPlcValue();
                                if(numValues == 1) {
                                    Object obj = response.getObject(fieldName);

                                    logger.info("Value[" + fieldName + "]: " + response.getObject(fieldName));
                                }
                                else {
                                    logger.info("Value[" + fieldName + "]:");
                                    for(int i = 0; i < numValues; i++) {
                                        logger.info(" - " + response.getObject(fieldName, i));
                                    }
                                }
                            }
                            else {
                                logger.error("Error[" + fieldName + "]: " + response.getResponseCode(fieldName).name());
                            }
                        }

                    }catch (Exception e){
                        logger.error(e.getMessage());
                    }
                }else {
                    System.out.println("断线后 不能 read");
                }
            }else {
                System.out.println("conn 为 connected");
            }

        }catch (Exception e){
            System.out.println(e.getMessage());
        }


        return R.ok().put("res",1);
    }

}
